@@ -15,6 +15,7 @@
 
 from __future__ import annotations
 
+from collections import defaultdict
 from collections.abc import Iterable
 import copy
 from dataclasses import dataclass, field, replace
@@ -79,7 +80,36 @@ logger = logging.getLogger(__name__)
 
 @dataclass
 class CircuitOperation:
-    """Specific quantum operation applied on a specific part of the QPU, e.g. in a quantum circuit."""
+    r"""Specific quantum operation applied on a specific part of the QPU, e.g. in a quantum circuit.
+
+    We currently support the following native operations for circuit execution:
+
+    ================ =========== ======================================= ===========
+    name             # of qubits args                                    description
+    ================ =========== ======================================= ===========
+    measure          >= 1        ``key: str``, ``feedback_key: str``     Measurement in the Z basis.
+    prx              1           ``angle: float``, ``phase: float``      Phased x-rotation gate.
+    cc_prx           1           ``angle: float``, ``phase: float``,
+                                 ``feedback_qubit: str``,
+                                 ``feedback_key: str``                   Classically controlled PRX gate.
+    reset            >= 1                                                Reset the qubit(s) to :math:`|0\rangle`.
+    cz               2                                                   Controlled-Z gate.
+    move             2                                                   Move a qubit state between a qubit and a
+                                                                         computational resonator, as long as
+                                                                         at least one of the components is
+                                                                         in the :math:`|0\rangle` state.
+    barrier          >= 1                                                Execution barrier.
+    delay            >= 1        ``duration: float``                     Force a delay between circuit operations.
+    ================ =========== ======================================= ===========
+
+    For each CircuitOperation you may also optionally specify :attr:`implementation`,
+    which contains the name of an implementation of the operation to use.
+    Support for multiple implementations is currently experimental and in normal use the
+    field should be omitted, this selects the default implementation for the operation for that locus.
+
+    See the submodules under :mod:`iqm.pulse.gates` for more details about each operation.
+
+    """
 
     name: str
     """name of the quantum operation"""
@@ -100,31 +130,54 @@ class CircuitOperation:
             ValueError: operation is not valid
 
         """
-        # find the op type
         op_type = op_table.get(self.name)
         if op_type is None:
-            raise ValueError(f"Unknown quantum operation '{self.name}'.")
+            message = ", ".join(op_table)
+            raise ValueError(f"Unknown operation '{self.name}'. Supported operations are '{message}'.")
+        self._validate_implementation(op_type)
+        self._validate_locus(op_type)
+        self._validate_args(op_type)
 
-        # find the implementation
-        impl_name = self.implementation
-        if impl_name is not None and impl_name not in op_type.implementations:
-            raise ValueError(f"Unknown implementation '{impl_name}' for quantum operation '{self.name}'.")
+    def _validate_implementation(self, op_type: QuantumOp) -> None:
+        if self.implementation is not None:
+            if not self.implementation:
+                raise ValueError("Implementation of the instruction should be None, or a non-empty string")
+            if self.implementation not in op_type.implementations:
+                raise ValueError(f"Unknown implementation '{self.implementation}' for quantum operation '{self.name}'.")
 
-        if 0 < op_type.arity != len(self.locus):
+    def _validate_locus(self, op_type: QuantumOp) -> None:
+        arity = op_type.arity
+        if (0 < arity) and (arity != len(self.locus)):
             raise ValueError(
-                f"The '{self.name}' operation acts on {op_type.arity} qubit(s), "
-                f"but {len(self.locus)} were given: {self.locus}"
+                f"The '{self.name}' operation acts on {arity} qubit(s), but {len(self.locus)} were given: {self.locus}."
             )
-
         if len(self.locus) != len(set(self.locus)):
             raise ValueError(f"Repeated locus components: {self.locus}.")
 
-        if not set(op_type.params).issubset(self.args):
+    def _validate_args(self, op_type: QuantumOp) -> None:
+        # Check argument names
+        submitted_arg_names = set(self.args)
+        allowed_arg_types = op_type.params | op_type.optional_params
+        if not set(op_type.params) <= submitted_arg_names:
             raise ValueError(
-                f"The '{self.name}' operation requires the arguments {op_type.params}, "
-                f"but {tuple(self.args)} were given."
+                f"The operation '{self.name}' requires "
+                f"the argument(s) {tuple(op_type.params)}, "
+                f"but {tuple(submitted_arg_names)} were given."
             )
 
+        if not submitted_arg_names <= set(allowed_arg_types):
+            allowed_arg_names = tuple(allowed_arg_types)
+            message = f"the arguments {allowed_arg_names}" if allowed_arg_names else "no arguments"
+            raise ValueError(f"The operation '{self.name}' allows {message}, but {submitted_arg_names} were given.")
+        # Check argument types
+        for arg_name, arg_value in self.args.items():
+            allowed_types = allowed_arg_types[arg_name]
+            if not isinstance(arg_value, allowed_types):
+                raise ValueError(
+                    f"The argument '{arg_name}' should be of one of the following supported types"
+                    f" {allowed_types}, but ({type(arg_value)}) was given."
+                )
+
 
 def load_config(path: str) -> tuple[QuantumOpTable, OpCalibrationDataTree]:
     """Load quantum operation definitions and calibration data from a YAML config file.
@@ -983,7 +1036,9 @@ class ScheduleBuilder:
             boxes.append(factory(**op.args))
         return TimeBox.composite(boxes, label=name, scheduling_algorithm=scheduling_algorithm)
 
-    def timeboxes_to_front_padded_playlist(self, boxes: Iterable[TimeBox], *, neighborhood: int = 0) -> Playlist:
+    def timeboxes_to_front_padded_playlist(
+        self, boxes: Iterable[TimeBox], *, neighborhood: int = 0
+    ) -> tuple[Playlist, dict[str, set[str]]]:
         """Temporary helper function, for converting a sequence of TimeBoxes to a Playlist.
 
         Each individual TimeBox in ``boxes`` is resolved into a Schedule, and then
@@ -1002,7 +1057,8 @@ class ScheduleBuilder:
                 blocks only the defined locus components and any other components which have occupied channels.
 
         Returns:
-            playlist that implements ``boxes``
+            playlist that implements ``boxes`` and the mapping from readout labels to the set of used measurement
+            implementations (of the format ``<QuantumOp name>.<impl name>``).
 
         """
         schedules = [self.resolve_timebox(box, neighborhood=neighborhood).cleanup() for box in boxes]
@@ -1046,7 +1102,7 @@ class ScheduleBuilder:
             playlist that implements ``boxes``
 
         """
-        return self.build_playlist(self.timebox_to_schedule(box, neighborhood=neighborhood) for box in boxes)
+        return self.build_playlist(self.timebox_to_schedule(box, neighborhood=neighborhood) for box in boxes)[0]
 
     def timebox_to_schedule(
         self,
@@ -1396,7 +1452,9 @@ class ScheduleBuilder:
             if not schedule[ch]:
                 schedule.append(ch, Block(T))
 
-    def build_playlist(self, schedules: Iterable[Schedule], finish_schedules: bool = True) -> Playlist:
+    def build_playlist(  # noqa: PLR0915
+        self, schedules: Iterable[Schedule], finish_schedules: bool = True
+    ) -> tuple[Playlist, dict[str, set[str]]]:
         """Build a playlist from a number of instruction schedules.
 
         This involves compressing the schedules so that no duplicate information
@@ -1410,7 +1468,8 @@ class ScheduleBuilder:
                 unless some process has already finalised them before calling this function.
 
         Returns:
-            playlist containing the schedules
+            playlist containing the schedules and the mapping from readout labels to the set of used measurement
+            implementations (of the format ``<QuantumOp name>.<impl name>``).
 
         Raises:
             ValueError: if the schedules contain channels with non-uniform sampling rates
@@ -1458,6 +1517,75 @@ class ScheduleBuilder:
 
                 # add the schedules in the playlist
 
+        label_to_implementation: dict[str, set[str]] = defaultdict(set)
+
+        def _map_instruction(inst: Instruction) -> sc_instructions.Instruction:
+            """TODO only necessary until SC has been updated to use the iqm.pulse Instruction class."""
+            operation: Any
+
+            def _map_acquisition(acq: AcquisitionMethod) -> sc_instructions.AcquisitionMethod:
+                if acq.implementation is not None:
+                    label_to_implementation[acq.label].add(acq.implementation)
+                if isinstance(acq, ThresholdStateDiscrimination):
+                    return sc_instructions.ThresholdStateDiscrimination(
+                        label=acq.label,
+                        delay_samples=acq.delay_samples,
+                        weights=_map_instruction(acq.weights).operation,
+                        threshold=acq.threshold,
+                        feedback_signal_label=acq.feedback_signal_label,
+                    )
+                if isinstance(acq, ComplexIntegration):
+                    return sc_instructions.ComplexIntegration(
+                        label=acq.label,
+                        delay_samples=acq.delay_samples,
+                        weights=_map_instruction(acq.weights).operation,
+                    )
+                if isinstance(acq, TimeTrace):
+                    return sc_instructions.TimeTrace(
+                        label=acq.label, delay_samples=acq.delay_samples, duration_samples=acq.duration_samples
+                    )
+                raise ValueError(f"Unknown AcquisitionMethod {acq}")
+
+            if isinstance(inst, Wait):
+                operation = sc_instructions.Wait()
+            elif isinstance(inst, VirtualRZ):
+                operation = sc_instructions.VirtualRZ(inst.phase_increment)
+            elif isinstance(inst, RealPulse):
+                operation = sc_instructions.RealPulse(
+                    wave=to_canonical(inst.wave),
+                    scale=inst.scale,
+                )
+            elif isinstance(inst, IQPulse):
+                operation = sc_instructions.IQPulse(
+                    wave_i=to_canonical(inst.wave_i),
+                    wave_q=to_canonical(inst.wave_q),
+                    scale_i=inst.scale_i,
+                    scale_q=inst.scale_q,
+                    phase=inst.phase,
+                    modulation_frequency=inst.modulation_frequency,
+                    phase_increment=inst.phase_increment,
+                )
+            elif isinstance(inst, ConditionalInstruction):
+                if len(inst.outcomes) != 2:
+                    raise ValueError("ConditionalInstruction requires exactly two outcomes.")
+                operation = sc_instructions.ConditionalInstruction(
+                    condition=inst.condition,
+                    if_true=_map_instruction(inst.outcomes[1]),
+                    if_false=_map_instruction(inst.outcomes[0]),
+                )
+            elif isinstance(inst, MultiplexedIQPulse):
+                sc_entries = tuple((_map_instruction(p), d) for p, d in inst.entries)
+                operation = sc_instructions.MultiplexedIQPulse(sc_entries)
+            elif isinstance(inst, ReadoutTrigger):
+                sc_acquisitions = tuple(_map_acquisition(a) for a in inst.acquisitions)
+                operation = sc_instructions.ReadoutTrigger(
+                    probe_pulse=_map_instruction(inst.probe_pulse),
+                    acquisitions=sc_acquisitions,
+                )
+            else:
+                raise ValueError(f"{inst} not supported.")
+            return sc_instructions.Instruction(duration_samples=int(inst.duration), operation=operation)
+
         # NOTE that there is no implicit right-alignment or equal duration for schedules, unlike in old-style playlists!
         for schedule in schedules:
             finished_schedule = self._finish_schedule(schedule) if finish_schedules else schedule
@@ -1486,7 +1614,7 @@ class ScheduleBuilder:
                 if prev_wait:
                     _append_to_schedule(sc_schedule, channel_name, prev_wait)
             pl.segments.append(sc_schedule)
-        return pl
+        return pl, label_to_implementation
 
     def _set_gate_implementation_shortcut(self, op_name: str) -> None:
         """Create shortcut for `self.get_implementation(<op_name>, ...)` as `self.<op_name>(...)`.
@@ -1522,67 +1650,3 @@ class ScheduleBuilder:
                 f"a class attribute ``ScheduleBuilder.{op_name}``."
             )
             self._logger.warning(warning_msg)
-
-
-def _map_instruction(inst: Instruction) -> sc_instructions.Instruction:
-    """TODO only necessary until SC has been updated to use the iqm.pulse Instruction class."""
-    operation: Any
-
-    def _map_acquisition(acq: AcquisitionMethod) -> sc_instructions.AcquisitionMethod:
-        if isinstance(acq, ThresholdStateDiscrimination):
-            return sc_instructions.ThresholdStateDiscrimination(
-                label=acq.label,
-                delay_samples=acq.delay_samples,
-                weights=_map_instruction(acq.weights).operation,
-                threshold=acq.threshold,
-                feedback_signal_label=acq.feedback_signal_label,
-            )
-        if isinstance(acq, ComplexIntegration):
-            return sc_instructions.ComplexIntegration(
-                label=acq.label, delay_samples=acq.delay_samples, weights=_map_instruction(acq.weights).operation
-            )
-        if isinstance(acq, TimeTrace):
-            return sc_instructions.TimeTrace(
-                label=acq.label, delay_samples=acq.delay_samples, duration_samples=acq.duration_samples
-            )
-        raise ValueError(f"Unknown AcquisitionMethod {acq}")
-
-    if isinstance(inst, Wait):
-        operation = sc_instructions.Wait()
-    elif isinstance(inst, VirtualRZ):
-        operation = sc_instructions.VirtualRZ(inst.phase_increment)
-    elif isinstance(inst, RealPulse):
-        operation = sc_instructions.RealPulse(
-            wave=to_canonical(inst.wave),
-            scale=inst.scale,
-        )
-    elif isinstance(inst, IQPulse):
-        operation = sc_instructions.IQPulse(
-            wave_i=to_canonical(inst.wave_i),
-            wave_q=to_canonical(inst.wave_q),
-            scale_i=inst.scale_i,
-            scale_q=inst.scale_q,
-            phase=inst.phase,
-            modulation_frequency=inst.modulation_frequency,
-            phase_increment=inst.phase_increment,
-        )
-    elif isinstance(inst, ConditionalInstruction):
-        if len(inst.outcomes) != 2:
-            raise ValueError("ConditionalInstruction requires exactly two outcomes.")
-        operation = sc_instructions.ConditionalInstruction(
-            condition=inst.condition,
-            if_true=_map_instruction(inst.outcomes[1]),
-            if_false=_map_instruction(inst.outcomes[0]),
-        )
-    elif isinstance(inst, MultiplexedIQPulse):
-        sc_entries = tuple((_map_instruction(p), d) for p, d in inst.entries)
-        operation = sc_instructions.MultiplexedIQPulse(sc_entries)
-    elif isinstance(inst, ReadoutTrigger):
-        sc_acquisitions = tuple(_map_acquisition(a) for a in inst.acquisitions)
-        operation = sc_instructions.ReadoutTrigger(
-            probe_pulse=_map_instruction(inst.probe_pulse),
-            acquisitions=sc_acquisitions,
-        )
-    else:
-        raise ValueError(f"{inst} not supported.")
-    return sc_instructions.Instruction(duration_samples=int(inst.duration), operation=operation)
