@@ -516,6 +516,9 @@ class CustomIQWaveforms(GateImplementation):
                     if k not in cls.excluded_parameters
                 }
                 cls.parameters = root_parameters | {"i": parameters_i, "q": parameters_q}
+        # allow Mixins of CustomIQWaveforms and CompositeGate
+        if issubclass(cls, CompositeGate):
+            init_subclass_composite(cls)
 
 
 def get_waveform_parameters(wave: type[Waveform], label_prefix: str = "") -> dict[str, Setting | Parameter]:
@@ -599,6 +602,20 @@ class SinglePulseGate(GateImplementation):
         return self.builder.channels[self.channel].duration_to_seconds(self.pulse.duration)
 
 
+def init_subclass_composite(gate_class: type[CompositeGate]) -> None:
+    if not gate_class.registered_gates:
+        # this would be pointless
+        raise ValueError(f"CompositeGate {gate_class.__name__} has no registered gates.")
+    # TODO we should also check that customizable_gates may_have_calibration (otherwise it's pointless
+    # to call them customizable), but we don't currently have access to their implementation classes here...
+    if gate_class.customizable_gates is None:
+        gate_class.customizable_gates = gate_class.registered_gates
+    elif not set(gate_class.customizable_gates) <= set(gate_class.registered_gates):
+        raise ValueError(
+            f"CompositeGate {gate_class.__name__}: customizable_gates must be a subset of registered_gates."
+        )
+
+
 class CompositeGate(GateImplementation):
     """Base class for gate implementations that are defined in terms of other gate implementations.
 
@@ -649,15 +666,7 @@ class CompositeGate(GateImplementation):
     """
 
     def __init_subclass__(cls):
-        if not cls.registered_gates:
-            # this would be pointless
-            raise ValueError(f"CompositeGate {cls.__name__} has no registered gates.")
-        # TODO we should also check that customizable_gates may_have_calibration (otherwise it's pointless
-        # to call them customizable), but we don't currently have access to their implementation classes here...
-        if cls.customizable_gates is None:
-            cls.customizable_gates = cls.registered_gates
-        elif not set(cls.customizable_gates) <= set(cls.registered_gates):
-            raise ValueError(f"CompositeGate {cls.__name__}: customizable_gates must be a subset of registered_gates.")
+        init_subclass_composite(cls)
 
     @classmethod
     def optional_calibration_keys(cls) -> tuple[str, ...]:
