@@ -29,7 +29,14 @@ import numpy as np
 
 from exa.common.data.parameter import Parameter, Setting
 from exa.common.qcm_data.chip_topology import DEFAULT_2QB_MAPPING
-from iqm.pulse.gate_implementation import GateImplementation, Locus, OILCalibrationData, get_waveform_parameters
+from iqm.pulse.gate_implementation import (
+    CompositeGate,
+    GateImplementation,
+    Locus,
+    OILCalibrationData,
+    get_waveform_parameters,
+    init_subclass_composite,
+)
 from iqm.pulse.playlist.instructions import Block, FluxPulse, Instruction, IQPulse, VirtualRZ
 from iqm.pulse.playlist.schedule import Schedule
 from iqm.pulse.playlist.waveforms import (
@@ -200,6 +207,8 @@ class FluxPulseGate(GateImplementation):
             parameters["qubit"]["amplitude"] = Parameter("", "Qubit flux pulse amplitude", "")
 
         cls.parameters = root_parameters | {k: v for k, v in parameters.items() if k not in cls.excluded_parameters}
+        if issubclass(cls, CompositeGate):
+            init_subclass_composite(cls)
 
     def _call(self) -> TimeBox:
         timebox = self.to_timebox(self._schedule)
